#define _CRT_SECURE_NO_WARNINGS

// #define TRACE
#define DUMP

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#ifdef WIN32
#include <process.h>
#include <io.h>
#include <direct.h>

#define open _open
#define read _read
#define lseek _lseek
#define close _close

#define O_BINARY _O_BINARY
#define ctime _ctime
#else
#include <unistd.h>
#include <stdlib.h>

#define O_BINARY 0
#endif

#include "xmalloc.h"
#include "object.h"
#include "util.h"
#include "Attribute.h"
#include "AttributeCache.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "EntityInstance.h"
#include "RelationInstance.h"
#include "LandscapeClassObject.h"
#include "EntityClass.h"
#include "RelationClass.h"
#include "TAInput.h"
#include "boundaries.h"

static CTAInput 		ta;
static CEntityInstance	*g_headP = 0;
static int				g_lines_code     = 0;
static int				g_dead_functions = 0;
static int				g_stripped       = 0;
static int				g_dead_lines     = 0;
static int				g_changed        = 0;
static int				g_updated        = 0;

int						g_trace_flag     = 0;
int						g_debug_flag     = 0;

/* Reduce all references to a function (in a header file) seen in amy
 * source files to a single reference. Don't want to treat a function
 * in a header as dead code unless there really are no links to it
 * from any source file
 */

static void
consolidate(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP, **ePP, *e1P;
	const char			*labelP;
	CRelationInstance	*riP, *nextP;
	const char			*fullpathP;
	char				*P;
	int					diff;

	// Sort by source file name and then by lineno
	for (ecP = ta.getEntityClass(0); ecP; ecP = ecP->m_nextP) {
		labelP = ecP->getId();
		if (!labelP) {
			continue;
		}
		if (strcmp(labelP, "GF") && strcmp(labelP, "LF")) {
			continue;
		}

		for (eP = ecP->m_entityInstancesP; eP; eP = eP->m_nextInClassP) {
			fullpathP = eP->m_fullpathP;
			if (!fullpathP || !strcmp(fullpathP, ".")) {
				P = 0;
				eP->buildPath(&P, 0);
				P = (char *) Xmalloc(P - ((char *) 0) + 1);
				eP->m_fullpathP = P;
				eP->buildPath(&P, 1);
			}
			fullpathP = eP->m_fullpathP;

			for (ePP = &g_headP; (e1P = *ePP); ePP = &(e1P->m_nextInOrderP)) {
				diff = strcmp(fullpathP, e1P->m_fullpathP);
				if (diff < 0) {
					goto found;
				}
				if (diff) {
					continue;
				}
				if (eP->m_lineno < e1P->m_lineno) {
					goto found;
				}
				if (eP->m_lineno == e1P->m_lineno) {
					// if source is compiled more than once
					// The same function might be named multiple times
					// printf("Consolidating %d:%s/%s in %s\n", eP->m_lineno, eP->getLabel(), e1P->getLabel(), eP->m_fullpathP);
					for (riP = eP->m_srcRelListP; riP; riP = nextP) {
						nextP = riP->m_next_srcP;
						// Move edge to be sourced by earlier node
						riP->m_next_srcP   = e1P->m_srcRelListP;
						e1P->m_srcRelListP = riP;
						riP->m_srcP        = e1P;
					}
					for (riP = eP->m_dstRelListP; riP; riP = nextP) {
						nextP = riP->m_next_dstP;
						// Moce edge to target earlier node
						riP->m_next_dstP   = e1P->m_dstRelListP;
						e1P->m_dstRelListP = riP;
						riP->m_dstP        = e1P;
					}
					eP->m_srcRelListP = 0;
					eP->m_dstRelListP = 0;
					goto skip;
			}	}	
found:		eP->m_nextInOrderP = *ePP;
			*ePP               = eP;
skip: 		continue;
	}	}	
}

static CEntityInstance *
find_function(CEntityInstance *eP)
{
	CRelationInstance	*riP;
	CRelationClass		*rcP;
	const char			*labelP;

	for (riP = eP->m_dstRelListP; riP; riP = riP->m_next_dstP) {
		rcP    = riP->getRelationClass();
		labelP = rcP->getId();
		if (!strcmp(labelP, "contain")) {
			return(riP->m_srcP);
	}	}
	assert(0);
	return(0);
}

static void
visit(CEntityInstance *eP, int depth)
{
	CRelationInstance	*riP;
	CRelationClass		*rcP;
	const char			*labelP;
	CEntityInstance		*dstP;

	assert(!eP->m_visited);
	eP->m_visited = CalledX;
#ifdef TRACE
	printf("%p ", eP);
	for (int i = 0; i < depth; ++i) {
		fputs("  ", stdout);
	}
	printf("%s %d:%s\n", eP->getLabel(), eP->m_lineno, eP->m_fullpathP);
#endif

	for (riP = eP->m_srcRelListP; riP; riP = riP->m_next_srcP) {
		rcP    = riP->getRelationClass();
		labelP = rcP->getId();
		dstP   = riP->m_dstP;
		if (labelP[1] == 'E') {
			// External variable or function
			continue;
		}
		switch (labelP[0]) {
		case 'C':
			if (!strcmp(labelP, "CIF")) {
				dstP = find_function(dstP);
			} 
			break;
		} 
		if (!dstP->m_visited) {
			visit(dstP, depth+1);
}	}	}

static void
depth_search(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP;
	const char			*labelP;

	// Find the functions and variables not visited
	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
		labelP = eP->getLabel();
		if (!strcmp(labelP, "main") || !strcmp(labelP, "wmain")) {
			ecP    = eP->getEntityClass();
			labelP = ecP->getId();
			if (!labelP) {
				continue;
			}
			if (!strcmp(labelP, "GF")) {
				visit(eP, 0);
}	}	}	}

#ifndef SSIZE_MAX
#define SSIZE_MAX 0x7FFFFFFF
#endif

#define ReportX	1
#define ReadX	2
#define WriteX	4
#define UpdateX	8
#define StripX	16

static int g_mode = 0;

const char	*g_filenameP = "";
static char	*g_sourceP   = 0;
static int	g_source_lth = 0;

static void
strip_source(void)
{
	char 	*startP, *atP, *at1P;
	char	c;
	int		stripped;

	stripped = g_stripped;
	for (atP = at1P = startP = g_sourceP; ; ++atP) {
		switch ((c = *atP)) {
		case 0:
			goto done;
		case '#':
			if (!strncmp(atP, "#if !defined(_DEAD_)", 20)) {
				++g_stripped;
				for (;; ++atP) {
					if (!*atP) {
						goto done;
					}
					if (!strncmp(atP, "#endif", 6)) {
						break;
				}	}
				atP = strchr(atP, '\n');
				if (!atP) {
					goto done;
				}
				continue;
			}
			if (!strncmp(atP, "#ifndef _DEAD_", 13) ||
				!strncmp(atP, "#endif /* _DEAD_", 16)) {
				++g_stripped;
				if (at1P > startP && at1P[-1] == '\n') {
					--at1P;
				}
				atP = strchr(atP, '\n');
				if (!atP) {
					goto done;
				}
				continue;
		}	}
		*at1P++ = c;
	}
done:
	*at1P = 0;
	if (g_stripped != stripped) {
		g_changed = 1;
	}
}

static int
load_source(void)
{
	struct stat buf;
	int			ret, lth, segment;
	char		*atP, *endP;
	int			fileno, mode;

	if (g_mode & WriteX) {
		mode = O_RDONLY;
	} else {
		mode = O_RDWR;
	}

	fileno = open(g_filenameP, mode | O_BINARY, 0);
	if (fileno < 0) {
		fprintf(stderr, "** Can't open %s for reading", g_filenameP);
		if (mode == O_RDWR) {
			fprintf(stderr, " and update");
		}
		fprintf(stderr, "\n");
		perror("");
		return -1;
	}
	// We have to scan the input more than once so read the entire input
	// into memory

	ret = fstat(fileno, &buf);
	if (ret != 0) {
		fprintf(stderr,"** fstat %s returned %d\n", g_filenameP, ret);
		perror("");
		goto bad_source;
	}

	// Allocate a large enough memory buffer

	lth = buf.st_size;
	if (!lth) {
		fprintf(stderr, "** %s is empty\n", g_filenameP);
		goto bad_source;
	}

	if (g_source_lth <= lth) {
		Xfree(g_sourceP);
		g_sourceP     = (char *) Xmalloc(lth + 1);
		g_source_lth  = lth;
	}
	atP          = g_sourceP;
	
	// Read the source into memory

	for (endP = atP + lth; atP < endP; atP += ret) {
		segment = endP - atP;
		if (segment > SSIZE_MAX) {
			segment = SSIZE_MAX;
		}
		ret = read(fileno, atP, segment);
		if (ret <= 0) {
			fprintf(stderr, "** read %s returned %d\n", g_filenameP, ret);
			perror("");
bad_source:
			close(fileno);
			return -1;
	}	}
	assert(atP == endP);
	*endP = 0;
	
	close(fileno);

	if (g_mode & StripX) {
		strip_source();
	}

	return(0);
}

extern boundaryT	*g_boundariesP;
extern int			g_boundaries;

static void
update_source(void)
{
	char		filename[4096];
	const char	*filenameP;
	FILE		*F;
	int			c;
	char		*atP;
	int			lineno, character, comment;
	boundaryT	*boundaryP, *end_boundaryP;
	int			find_lineno, find_char, add_end;

	if (!g_changed) {
		return;
	}

	++g_updated;
	if (g_mode & WriteX) {
		sprintf(filename, "%s.dead", g_filenameP);
		filenameP = filename;
	} else {
		filenameP = g_filenameP;
	}
	F = fopen(filenameP, "wb");
	if (!F) {
		fprintf(stderr, "%s Can't write\n", filenameP);
		perror("");
		return;
	}
	rewind(F);

	atP = g_sourceP;

	if (g_mode & (WriteX | UpdateX)) {
		extern	char **g_argsPP;
		char		**argPP;
		char		*argP, *P, *nameP;

		time_t		time_now;
		char		*dateP;
		struct tm 	*tmP;
		char		timestamp[20];

 		time( &time_now );
 		tmP = gmtime(&time_now);
		sprintf(timestamp, "%04d%02d%02d%02d%02d%02d", 		
			tmP->tm_year+1900, tmP->tm_mon+1, tmP->tm_mday,
			tmP->tm_hour, tmP->tm_min, tmP->tm_sec);

		dateP = ctime( &time_now ); 
		dateP[24] = 0;

		fprintf(F, "#if !defined(_DEAD_) ");
		for (argPP = g_argsPP; (argP = *argPP); ++argPP) {
			if (argP[0] == '-' && (argP[1] == 'D' || argP[1] == 'U')) {
				fprintf(F, "&& ");
				nameP = argP + 2;
				P = strchr(nameP, '=');
				if (P) {
					*P = 0;
				}
				if (!P || P[1] == '"') {
					// Can't use strings in the conditional language
					// supported by CPP
					if (argP[1] == 'U') {
						fputc('!', F);
					}
					fprintf(F, "defined(%s) ", nameP);
				} else {
					fprintf(F, "defined(%s) && (%s == %s) ", nameP, nameP, P+1);
				}
				if (P) {
					*P = '=';
				}
		}	}
		fprintf(F, "\n#define _DEAD_%s\n#endif /* _DEAD_ */\n", timestamp);

		lineno        = 1;
		character     = 0;
		comment       = 0;
		boundaryP 	  = g_boundariesP;
		for (end_boundaryP = boundaryP + g_boundaries; boundaryP < end_boundaryP; ++boundaryP) {
			if (!boundaryP->dead || boundaryP->lineno[1] <= 0) {
				continue;
			}
			for (add_end = 0; add_end < 2; ++add_end) {
				find_lineno = boundaryP->lineno[add_end];
				find_char   = boundaryP->character[add_end];
				if (lineno > find_lineno) {
					goto skip;
				}
				for (; (c = *atP); ) {
					switch (comment) {
					case 0:
						if (c == '/') {
							if (atP[1] == '*') {
								fputc(c, F);
								c       = *(++atP);
								comment = 1;
								break;
							}
							if (atP[1] == '/') {
								comment = 2;
								break;
						}	}
						++character;
						break;
					case 1:
						if (c == '*' && atP[1] == '/') {
							fputs("*/", F);
							atP    += 2;
							comment = 0;
							continue;
						}
						break;
					case 2:
						if (c == '\n') {
							comment = 0;
						}
						break;
					}
							
					if (lineno == find_lineno && character == find_char) {
						if (!add_end) {
							fprintf(F, "\n#ifndef _DEAD_%s /* %s */\n", timestamp, dateP);
						} else {
							fprintf(F, "\n#endif /* _DEAD_%s %s */\n", timestamp, dateP);
							g_lines_code += find_lineno - boundaryP->lineno[0] + 1;
						}
						--character;
						break;
					}
					fputc(c, F);
					++atP;
					if (c == '\n') {
						++lineno;
						if (lineno > find_lineno) {
							goto skip;
						}
						character = 0;
			}	}	}
skip:		continue;
	}	}

	for (; (c = *atP); ++atP) {
		fputc(c, F);
	}
	fclose(F);
}

static void
report_dead(void)
{
	CEntityClass		*ecP;
	CEntityInstance		*eP;
	const char			*labelP;
	const char			*typeP;
	int					load_failed = 1;
	int					boundaries  = -1;
	int					lineno;
	boundaryT			*boundaryP, *end_boundaryP;

	for (eP = g_headP; eP; eP = eP->m_nextInOrderP) {
		if (eP->m_visited) {
			continue;
		}
		if (strcmp(g_filenameP, eP->m_fullpathP)) {
			/* Commit update to last source */

			update_source();
			g_changed   = 0;
			g_filenameP = eP->m_fullpathP;
			load_failed = 1;
			boundaries  = -1;
			if (g_mode & ReadX) {
				load_failed = load_source();
				if (load_failed) {
					fprintf(stderr, "%s Can't read\n", g_filenameP);
			}	}
			if (g_mode & ReportX) {
				printf("%s\n", g_filenameP);
		}	}
		lineno = eP->m_lineno;
		++g_dead_functions;
		if (g_mode & ReportX) {
			ecP    = (CEntityClass *) eP->getParentClass();
			typeP  = ecP->getLabel();
			labelP = eP->getLabel();
			if (lineno == NoLineno) {
				printf("\t");
			} else {
				printf("\t%d", lineno);
			}
			printf("\t%s %s not used\n", typeP, labelP);
		}
		if (!load_failed && lineno != NoLineno && (g_mode & (WriteX | UpdateX))) {
			switch (boundaries) {
			case -1:
				boundaries = establish_boundaries(eP);
				if (boundaries > 0) {
					break;
				}
			case 0:
				continue;
			}

			boundaryP = g_boundariesP;
			for (end_boundaryP = boundaryP + boundaries; ; ++boundaryP) {
				if (end_boundaryP <= boundaryP) {
					fprintf(stderr, "Couldn't find boundaries for %s:%s:%d\n", g_filenameP, eP->getLabel(), lineno);
					break;
				} 
				if (boundaryP->lineno[0] <= lineno && boundaryP->lineno[1] >= lineno) {
					boundaryP->dead = 1;
					g_changed       = 1;
					break;
	}	}	}	}

	update_source();
}

static void
find_dead(const char *filenameP)
{
	ta.loadTA(filenameP);
	consolidate();
	depth_search();
	report_dead();
}

int 
main(int argc, const char *argv[])
{
	const char *filenameP = 0;
	int			i;
	const char	*argP;

	printf("%s " VERSION " " __DATE__ " " __TIME__ "\n", argv[0]);
	for (i = 1; i < argc; ++i) {
		argP = argv[i];
		switch (argP[0]) {
		case '-':
			switch (argP[1]) {
			case 'd':
				g_debug_flag = 1;
				break;
			case 'r':
				g_mode |= ReadX;
				break;
			case 'w':
				g_mode |= ReadX | WriteX;
				break;
			case 'v':
				g_mode |= ReportX;
				break;
			case 'u':
				g_mode |= ReadX | UpdateX;
				break;
			case 's':
				g_mode |= ReadX | StripX;
				break;
			case 't':
				g_trace_flag = 1;
				break;
			default:
				fprintf(stderr, "Unknown option %s\n", argP);
				fprintf(stderr, "usage: %s [-r/ead] [-s/trip] [-w/rite|-u/pdate] [-v/erbose] [-t/race] [<] <tafile>\n", argv[0]);
				return(1);
			}
			break;
		default:
			if (filenameP) {
				fprintf(stderr, "Multiple ta_files\n");
				fprintf(stderr, "usage: %s [<] <tafile>\n", argv[0]);
				return(1);
			}
			filenameP = argP;
			break;
	}	}
	if (!g_mode) {
		g_mode = ReportX;
	}
	find_dead(filenameP);
	printf("%d lines of code #ifdef'd out\n", g_lines_code);
	if (g_stripped) {
		printf("%d #ifdef _DEAD_ statements removed\n", g_stripped);
	}
	printf("%d functions are dead\n", g_dead_functions);
	if (g_dead_lines) {
		printf("%d lines of source code marked dead\n", g_dead_lines);
	}
	if (g_updated) {
		printf("%d files updated\n", g_updated);
	}
	return 0;
}

